module DataHelpers
  extend ActiveSupport::Concern

  module JobStates
    FINISHED = 'FINISHED'
    STARTED = 'STARTED'
    ABORTED = 'ABORTED'
  end

  MACHINE_STATES = [:on, :notready, :ready, :waitingforjobs, :processing, :maintenance, :error, :halt, :estop, :setup, :off]

  module MachineStates
    ON = 0
    NOTREADY = 1
    READY = 2
    WAITINGFORJOBS = 3
    PROCESSING = 4
    MAINTENANCE = 5
    ERROR = 6
    HALT = 7
    ESTOP = 8
    SETUP = 9
    OFF = 10
  end

  JOB_STATES = {started: 3, aborted: 4, finished: 5, postponed: 6, skipped: 7, restarted: 8}
  JOB_STATES_INVERTED = JOB_STATES.invert

  ROUTING_KICKOUT_REASONS = {
    noValidPath: {database_key: 'no_valid_path'},
    routedWhileWaitingOtherProcess: {database_key: 'routed_while_waiting_other_process'},
    failFetchProcesses: {database_key: 'fail_fetch_processes'},
    markedDead: {database_key: 'marked_dead'},
    noProcesses: {database_key: 'no_processes'},
    receivedInvalidProcess: {database_key: 'received_invalid_process'},
    noProcessesAtNonStart: {database_key: 'no_processes_at_non_start'},
    manualReject: {database_key: 'manual_reject'},
    processAborted: {database_key: 'process_aborted'},
    timeSyncError: {database_key: 'time_sync_error'},
    fetchJobInfFailed: {database_key: 'fetch_job_inf_failed'},
    failedToSynchronize: {database_key: 'failed_to_synchronize'},
    returnJobToTray: {database_key: 'rework_complete'},
    failToScan: {database_key: 'other'},
    receivedStartInReject: {database_key: 'other'},
    completedInvalidProcessInDefault: {database_key: 'other'},
    other: {database_key: 'other'}
  }

  RETRIEVE_KICKOUT_REASONS = {
    noValidPath: {database_key: 'no_valid_path'},
    routedWhileWaitingOtherProcess: {database_key: 'routed_while_waiting_other_process'},
    failFetchProcesses: {database_key: 'fail_fetch_processes'},
    markedDead: {database_key: 'marked_dead'},
    noProcesses: {database_key: 'no_processes'},
    receivedInvalidProcess: {database_key: 'received_invalid_process'},
    noProcessesAtNonStart: {database_key: 'no_processes_at_non_start'},
    manualReject: {database_key: 'manual_reject'},
    processAborted: {database_key: 'process_aborted'},
    timeSyncError: {database_key: 'time_sync_error'},
    fetchJobInfFailed: {database_key: 'fetch_job_inf_failed'},
    failedToSynchronize: {database_key: 'failed_to_synchronize'},
    returnJobToTray: {database_key: 'rework_complete'},
    other: {database_key: 'other'}
  }

  def routing_kickout_reason_database_keys
    h = {}
    ROUTING_KICKOUT_REASONS.each do |k,v|
      h[v[:database_key]] = k
    end
    h
  end

  def routing_kickout_database_keys_arr
    routing_kickout_reason_database_keys.stringify_keys.keys
  end

  # Never trust parameters from the scary internet, only allow the white list through.
  def job_state_add_params(params)
    params[:job_state] = params[:jobstate] if params[:jobstate]
    job_state = params.is_a?(Hash) ? params[:job_state] : params.require(:job_state)

    machine = Machine.find_or_create_by(sn: params[:sn])

    # If there is no job number in the payload, use the batchId instead
    # Note, this solution is only to be applied until batching of jobs is properly implemented (http://omics.dyndns.biz:4747/rx-suite/modules/optional/data/-/issues/40)
    side_num = Sideable.side_letter_to_side_num(job_state[:lensside])
    job_number = job_state[:job] || job_state[:batchId]
    job = FindHelpers::JobFinders.find_active_or_create_by_number_and_side_num(job_number, side_num)

    # Convert the state of the job state to a symbol so that it can be used in the enum hash
    symbol_job_state = job_state_to_symbol(job_state[:state])

    state_of_job_state = job_state[:state] ? JOB_STATES[symbol_job_state] : nil

    {
      job: job,
      machine: machine,
      batch_id: job_state[:batchId],
      process: job_state[:process],
      substate: job_state[:substate],
      side: side_num,
      state: state_of_job_state,
      comment: job_state[:comment],
      scrap: job_state[:scrap],
      scrap_reason: job_state[:scrapReason],
      time: "#{params[:date]} #{params[:time]}"
    }
  end

  def post_job_state(sn, job_number, side, state, process, wait: nil)
    begin
      formattedParams = job_state_add_params({
                                               sn: sn,
                                               date: Time.now.to_date,
                                               time: Time.now.to_time,
                                               job_state: {
                                                 job: job_number,
                                                 lensside: side,
                                                 state: state,
                                                 process: process,
                                               }
                                             })

      if wait
        AddJobState.set(wait: wait).perform_later(formattedParams)
      else
        AddJobState.perform_later(formattedParams)
      end
    rescue => e
      Rails.logger.error e
    end
  end

  private

  def job_state_to_symbol(job_state)
    symbol_job_state = job_state
    if symbol_job_state.is_a? String
      downcase_string = job_state.downcase
      symbol_job_state = downcase_string.to_sym
    end
    symbol_job_state
  end
end
