class MachineMultiKpiTableDashboard < RxModule::Dashboards::ScoreCard
  include MachineDashboard
  include DataHelpers
  include MachineConcern
  include KpiHelpers
  include BucketCalculator

  class InvalidRowType < StandardError; end

  class << self
    def run_in_background?
      true
    end
  end

  expose('machine_multi_kpi_table', 'Machine Multi KPI Table', 'View one or more machine KPI in a table format.')
  parameter(:buckets, "Hours to display", "How many hours to display", :integer, required: true, default: 8, min: 3, max: 24)
  parameter(:offset, "Start at mid-hour?", "When enabled, column starts with start at the middle of the hour (ex: 1:30, 2:30, etc.)", :boolean, default: false)
  parameter(:row_type, "Row Type", "Type of calculation to use for this row?",:string, value_list: -> { row_type_array }, single_row_parameter: true)
  parameter(:row_name, "Row Name", "Name that will used in the first column of the row.", :string, single_row_parameter: true)
  targets(nil,"Machine", "The machine(s) to track", -> { machine_targets }, single_row_parameter: true, custom_targets: false)
  parameter(:data_operation, "Operation", "Complex operation to be done on each cell of the row", :string, single_row_parameter: true, required: false, value_list: -> { data_operations.to_a.map { |name, hash| hash.slice(:label, :value) }},
            sub_parameters: [
              {
                condition: "PERCENTAGE_OF_GOAL",
                parameter: sub_parameter(:target_number, "Target Number", "The target that the percentage will be calculated off of.", :integer, single_row_parameter: true, required: false)
              }
            ])

  def build_results!
    unroll_row_targets! # Unpack the machine groups so that all row targets are machine ids
    setup_targets! # Since one machine can appear in several rows
    setup_series!
  end

  def buckets
    @parameters[:buckets]
  end

  def offset
    @parameters[:offset]
  end

  def parameters
    @parameters
  end

  def rows
    @parameters[:rows]
  end

  def unroll_row_targets!
    rows.each do |row|
      target_ids = row['targets']
      next if target_ids.blank?

      row['targets'] = []
      target_ids.each do |id|
        category, ref = id.split("-")
        case category
        when 'm'
          row['targets'] << ref
        when 'g'
          g = MachineGroup.find(ref)
          g.machine_ids.each {|m_id| row['targets'] << m_id } unless g.nil?
        end
      end
    end
  end

  def setup_targets!
    ids = rows.map { |r| r['targets'] }.flatten
    @_machines = Machine.where(id: ids).find_each.to_a
  end

  def setup_series!
    @start_time, @end_time = dashboard_period

    rows.each do |row|
      row_type = row['row_type']

      if row_type == 'divider'
        add_row([])
        next
      end

      machine_ids = resolve_machine_ids(row['targets'])

      next if machine_ids.nil?

      options = {as_percent: false }

      case row_type
      when 'throughput'
        data = MachineKpi.throughput_between(@start_time, @end_time, Timeframe::HOUR_BY_HOUR, machine_ids)
        options[:as_percent] = row_op_is_percent_target?(row)
      when 'machine_aborts'
        data = MachineKpi.aborted_job_ratio_between(@start_time, @end_time, Timeframe::HOUR_BY_HOUR, machine_ids)
        options[:as_percent] = true
      when 'machine_uptime'
        data = MachineKpi.uptime_between(@start_time, @end_time, Timeframe::HOUR_BY_HOUR, machine_ids, options: {:as_percent => true})
        options[:as_percent] = true
      when 'machine_utilization'
        data = MachineKpi.utilization_between(@start_time, @end_time, Timeframe::HOUR_BY_HOUR, machine_ids, options: {:as_percent => true})
        options[:as_percent] = true
      else
        raise InvalidRowType
      end

      data = apply_row_operation(data, row)

      add_row(build_row(row['row_name'], format_data_buckets(data, options)))
    end
  end

  def row_op_is_percent_target?(row)
    return false if row['data_operation'].nil?
    row['data_operation'] == 'PERCENTAGE_OF_GOAL'
  end

  def apply_row_operation(data, row)
    if row_op_is_percent_target?(row)
      target = row['target_number'].to_f
      data.each { |k,v| data[k] = (v / target) * 100}
    end
    data
  end

  def machines
    @_machines ||= []
  end

  def row_targets(targets)
    return if targets.nil?
    targets.map { |t| @_machines.find { |m| m.id == t.to_i } }.compact
  end

  def resolve_machine_ids(targets)
    row_targets(targets).pluck(:id)
  end

  def columns
    # Redefine columns so that our columns are the buckets.
    c = time_buckets(@start_time, @end_time, Timeframe::HOUR_BY_HOUR).reverse_each.with_index.map do |(column_name, value), i|
      {id: "bucket_#{i}", name: column_formatting(column_name), type: 'time'}
    end
    c.unshift({id: "row_name", name: "Last run: #{Time.now.in_time_zone(ENV['TZ']).strftime("%H:%M:%S")}", type: 'string'})
  end

  def column_formatting(time)
    time.strftime("%H:%M")
  end

  def build_row(name, data)
    data.merge({row_name: name}.stringify_keys)
  end

  def format_data_buckets(data, options)
    formatted = {}
    data.reverse_each.with_index do |(k,v), i|
      formatted["bucket_#{i}"] = options[:as_percent] ? "#{v.round(1)}%" : v
    end
    formatted
  end

  # Calculate the start and end time of the dashboard. The start time and end time depend on if we're using an offset.
  # If we're using an offset, we start the first bucket at the mid-hour (ex: 4h30), otherwise, use the hour start (4h00)
  # With an offset, if we're closer to the start of the hour, then return the mid-hour, otherwise return the end of the hour.
  # For example, if it's 3h03 the bucket end is 3h30, however, if it's 3h37 the bucket end is 4h00.
  def dashboard_period
    if offset
      diff_from_hour_start = Time.zone.now - Time.zone.now.at_beginning_of_hour
      end_time = diff_from_hour_start <= KpiHelpers.cache_entry_size_seconds ? Time.zone.now.at_beginning_of_hour : Time.zone.now.at_beginning_of_hour+1.hour
      end_time = end_time.advance(minutes: 29, seconds: 59)
    else
      end_time = Time.zone.now.at_beginning_of_hour.advance(minutes: 59, seconds: 59)
    end

    start_time = end_time - (buckets*1.hour)+ 1.seconds
    [start_time, end_time]
  end

  class << self
    def target_to_name(target)
      if target.is_a? Machine
        "#{target.name} (#{target.sn})"
      elsif target.is_a? MachineGroup
        target.name
      end
    end

    def target_to_id(target)
      if target.is_a? Machine
        "m-#{target.id}"
      elsif target.is_a? MachineGroup
        "g-#{target.id}"
      end
    end

    def machine_targets
      MachineGroup.all.map do |g|
        { label: target_to_name(g), value: target_to_id(g), values_in_group: g.machines.map {|m| target_to_id(m)}, group: "Machine Group", description: "Report on percent uptime for the entire #{g.name} group" }
      end + Machine.all.map do |m|
        { label: target_to_name(m), value: target_to_id(m), group: "Machine", description: "Report on percent uptime for machine #{m.name}, sn #{m.sn}" }
      end
    end

    def row_types
      {
        throughput: {
          label: 'Throughput',
          value: 'throughput',
          description: 'Number of jobs completed by selected machine(s)',
          group: 'operational'
        },
        machine_kickout: {
          label: 'Aborted Job %',
          value: 'machine_aborts',
          description: 'Average % of jobs aborted by selected machine(s)',
          group: 'operational'
        },
        utilization: {
          label: 'Utilization %',
          value: 'machine_utilization',
          description: 'Average utilization % of selected machine(s)',
          group: 'operational'
        },
        uptime: {
          label: 'Uptime %',
          value: 'machine_uptime',
          description: 'Average uptime % of selected machine(s)',
          group: 'operational'
        },
        divider: {
          label: 'Divider',
          value: 'divider',
          description: 'Makes a small separation row in the table',
          group: 'display'
        }
      }
    end

    def data_operations
      {
        "PERCENTAGE_OF_GOAL" => {
          label: "% of target",
          value: "PERCENTAGE_OF_GOAL",
          operation_method: lambda { |val, goal| return "#{((val/[1.0,goal.to_f].max)*100).round(1)}%" },
          method_parameters: lambda { |parameters| return parameters["target_number"] }
        }
      }
    end

    def row_type_array
      row_types.values
    end
  end
end