import * as React from "react";
import {observer} from "mobx-react-lite";
import {Column, Filter, SortingRule} from "react-table";
import {MachineModel, MachineStore} from "../../stores/MachineStore";
import {TabPanelProps} from "./TabPanel";
import {period, PeriodFilterButtons} from "./PeriodFilterButtons";
import {DataStore, errToString, Model} from "@rx-suite/data";
import {ButtonField, DateRangeField, ModelTable} from "@rx-suite/ui";
import {Component, useState} from "react";
import {useSnackbar} from "notistack";
import {ImportExport, Info} from "@material-ui/icons";
import {t} from "@lingui/macro";
import {
    Dialog,
    DialogContent,
    DialogTitle,
    makeStyles,
    Typography,
    CircularProgress,
    Button,
    IconButton, Grid,
} from "@material-ui/core";
import moment from "moment";
import FileSaver from "file-saver";
import CloseIcon from "@material-ui/icons/Close";
import {MachineTypeStore} from "../../stores/MachineTypeStore";
import {grey} from "@material-ui/core/colors";

type PanelProps = {
    machine: MachineModel;
    enabledKey: string;
    tabName: string;
    columns: Column[];
    store: DataStore<Model>;
    subComponent?: (original: Model) => Component | undefined;
    defaultSorted?: SortingRule;
    defaultFiltered?: Filter[];
    exportedFileName?: string;
    exportCallback?: (startDate: string, endDate: string) => Promise<CSVExport | CSVError>;
} & TabPanelProps

type CSVExport = {
    csv: string;
}

type CSVError = {
    error: string;
};

const useStyles = makeStyles((theme) => ({
    buttonRow: {
        display: 'flex',
        flexDirection: 'row',
        justifyContent: 'space-between',
        alignItems: 'flex-start'
    },
    dialog: {
        margin: 0,
        padding: theme.spacing(2)
    },
    closeButton: {
        position: 'absolute',
        right: theme.spacing(1),
        top: theme.spacing(1),
        color: theme.palette.grey[500],
    },
    dialogContent: {
        padding: theme.spacing(2)
    },
    dialogHeader: {
        display: 'flex',
        flexDirection: 'row',
        justifyContent: 'space-between',
        alignItems: 'center'
    },
    dialogHeaderText: {
        margin: '0 auto'
    },
    exitButton: {
        position: 'absolute',
        top: '5px',
        right: '5px'
    },
    exportBtn: {
        marginTop: theme.spacing(2),
    },
    exportBtnSection: {
        display: 'flex',
        flexDirection: 'column',
        justifyContent: 'center',
        alignItems: 'center',
        marginTop: theme.spacing(1),
        marginBottom: theme.spacing(1)
    },
    csvButton: {
        display: 'flex',
        flexDirection: 'row',
        justifyContent: 'flex-end'
    },
    progressBar: {
        display: 'flex',
        flexDirection: 'column',
        justifyContent: 'center',
        alignItems: 'center',
        marginTop: theme.spacing(1),
        marginBottom: theme.spacing(1)
    },
    generatingText: {
        marginTop: theme.spacing(1)
    },
    infoIcon: {
        marginRight: theme.spacing(1),
        color: grey[600]
    },
    infoMessageWrapper: {
        marginLeft: theme.spacing(1),
        textAlign: 'center',
        color: grey[600],
        margin: `${theme.spacing(4)}px 0`,
    }
}));

export const MachineDataViewTab: React.FunctionComponent<PanelProps> = observer((props) => {
    const {machine, columns, store, subComponent, defaultSorted, defaultFiltered, exportedFileName, exportCallback, enabledKey, tabName} = props;

    const types = MachineTypeStore.getAll();


    const {enqueueSnackbar} = useSnackbar();

    const classes = useStyles();

    const [periodFilterNum, setPeriodFilterNum] = React.useState<number>(0);
    const [open, setOpen] = useState<boolean>(false);

    const [startDate, setStartDate] = useState<moment.Moment>(
        moment().startOf('day')
    );

    const [endDate, setEndDate] = useState<moment.Moment>(
        moment().endOf('day')
    );

    const archiveDate = moment(new Date()).subtract(32, 'd');

    const [generating, setGenerating] = useState<boolean>(false);
    const [exportBtn, setExportBtn] = useState<boolean>(false);
    const [csv, setCSV] = useState<string|null>(null);

    const handleClose = () => {
        if (generating) {
            return; // as long as an export is being generated, do not close the dialog
        }
        setOpen(false);
    };

    const handleExited = () => {
        setStartDate(moment().startOf('day'));
        setEndDate(moment().endOf('day'));
        setGenerating(false);
        setExportBtn(false);
        setCSV(null);
    };

    const openDialog = () => {
        setOpen(true);
    };

    const generateCSV = async () => {
        try {
            if (!exportCallback) {
                throw new Error(t`Export not provided for this data.`);
            }
            setGenerating(true);
            const exportedData = await exportCallback(formatToMillisecondUTC(startDate), formatToMillisecondUTC(endDate));

            if ('error' in exportedData) {
                throw new Error(t`Export failed.`);
            }
            setCSV(exportedData.csv);
            setGenerating(false);
            setExportBtn(true);

        } catch (e) {
            handleClose();
            enqueueSnackbar(errToString(e), {variant: 'error'});
        }
    };

    const downloadCSV = () => {
        try {
            if (!csv) {
                throw new Error(t`Export could not be downloaded.`);
            }
            FileSaver.saveAs(new Blob(
                    [csv],
                    {type: "text/csv;charset=utf-8"}),
                fileName()
            );
        } catch(e) {
            handleClose();
            enqueueSnackbar(errToString(e), {variant: 'error'});
        }
    };

    const fileName = () => {
        const base = `${machine.sn}-${formatToISODate(startDate)}-${formatToISODate(endDate)}`
        if(exportedFileName) {
            return `${base}-${exportedFileName}`
        } else {
            return `${base}-exported.csv`
        }
    };

    const formatToISODate = (date: moment.Moment) => {
        return date.format('YYYY_MM_DDTHHmmss');
    };

    const formatToMillisecondUTC = (date: moment.Moment) => {
        return date.utc().format('YYYY-MM-DDTHH:mm:ss.SSSZ');
    };

    let exportMsg = t`Generating CSV… This may take a few minutes to generate.`
    if (moment(startDate).isSameOrBefore(archiveDate) || moment(endDate).isSameOrBefore(archiveDate)) {
        exportMsg = t`Generating CSV… Data export is in archive. This may take longer than usual to generate.`
    }

    const dialogContent = <>
        <DialogTitle className={classes.dialogHeader} disableTypography>
            <Typography variant="h5" className={classes.dialogHeaderText}>
                {t`Export`}
            </Typography>
            <IconButton
                className={classes.exitButton}
                size={"small"}
                onClick={handleClose}
            >
                <CloseIcon />
            </IconButton>
        </DialogTitle>
        <DialogContent className={classes.dialogContent} dividers>
            {generating && <div className={classes.progressBar}>
                <CircularProgress/>
                <Typography className={classes.generatingText} variant="subtitle2" color="textSecondary">
                    {exportMsg}
                </Typography>
            </div>}
            {!generating && !exportBtn && <>
                <DateRangeField
                    startDate={startDate}
                    endDate={endDate}
                    onClose={() => {}}
                    onStartDateChange={(val) => {

                        if (typeof val === "string") return; // input was invalid

                        if (val.isAfter(endDate)) return; // cannot have startDate > endDate

                        setStartDate(val.set({ second: 0, millisecond: 0 }));
                    }}
                    onEndDateChange={(val) => {

                        if (typeof val === "string") return; // input was invalid

                        if (val.isBefore(startDate)) return; // cannot have endDate < startDate

                        setEndDate(val.set({ second: 59, millisecond: 999 }));
                    }}
                />
                <div className={classes.csvButton}>
                    <ButtonField
                        onClick={generateCSV}
                        text={t`Export CSV`}
                        startIcon={<ImportExport/>}
                        loading={generating}
                        tooltip={t`Export the contents of the table into a csv`}
                        disabled={!exportCallback}
                    />
                </div>
            </>}
            {!generating && exportBtn && <div className={classes.exportBtnSection}>
                <Typography variant="h6">
                    {t`CSV Successfully Generated.`}
                </Typography>
                <Button
                    className={classes.exportBtn}
                    color="primary"
                    variant="contained"
                    onClick={downloadCSV}
                >
                    {t`Download CSV`}
                </Button>
            </div>}
        </DialogContent>
    </>;

    const table = () => {
        const sorted = defaultSorted ?
            defaultSorted :
            [{id: "time", desc: true}]

        const filtered = defaultFiltered
            ? defaultFiltered.slice()
            : []
        ;

        // Adding a push with the period and machine sn since all filters will have it, might as well not require it from custom forms
        filtered.push({id: "sn", value: machine.sn});
        filtered.push({id: "period", value: period[periodFilterNum].key});

        return (
            <div>
                <ModelTable
                    suppressWrapper
                    store={store}
                    columns={columns}
                    pageSize={100}
                    SubComponent={subComponent ? ({original}) => { return subComponent(original); } : undefined}
                    defaultSorted={sorted}
                    defaultFiltered={filtered}
                    replacementActions={{
                        Delete: () => null
                    }}
                />
            </div>
        )
    }

    const machineType = React.useMemo(() => {
        return MachineTypeStore.models.find(type => type.id == machine.machine_type_id)
    }, [MachineTypeStore.models]);

    if(machineType && !machineType[enabledKey]) {
        return  <div>
                    <Grid container className={classes.infoMessageWrapper} alignItems={'center'} justifyContent={'center'}>
                        <Info fontSize={'medium'} className={classes.infoIcon}/><Typography variant={'h6'}>{t`This machine does not collect ${tabName} data. Please contact support for more information.`}</Typography>
                    </Grid>
                </div>
    }

    return (<>
        <div className={classes.buttonRow}>
            <PeriodFilterButtons value={periodFilterNum} change={setPeriodFilterNum} />
            {exportCallback &&
                <ButtonField
                    onClick={() => { openDialog() } }
                    text={t`Export`}
                    startIcon={<ImportExport/>}
                    loading={store.isLoading}
                    tooltip={t`Export the contents of the table`}
                    disabled={!exportCallback}
                />
            }
        </div>
            {table()}
        <Dialog
            className={classes.dialog}
            open={open}
            onClose={handleClose}
            TransitionProps={{
                onExited: handleExited
            }}
            maxWidth='xs'
            fullWidth
        >
            {dialogContent}
        </Dialog>
    </>);
});
