import * as React from "react";
import {observer} from "mobx-react-lite";
import {t} from "@lingui/macro";
import {TabPanelProps} from "../TabPanel"
import {MachineModel} from "../../../stores/MachineStore";
import {MachineDataViewTab} from "../MachineDataViewTab";
import {ErrorModel, ErrorStore} from "../../../stores/ErrorStore";
import {JobModel} from "../../../stores/JobStore";
import {makeColumn, SelectField} from "@rx-suite/ui";
import {Filter, ReactTableFunction} from "react-table";
import {getDateTimeString} from "../MachineDataView";
import {Moment} from "moment";

type PanelProps = {
    machine: MachineModel;
} & TabPanelProps

export const ErrorsPanel: React.FunctionComponent<PanelProps> = observer((props) => {
    const {machine} = props;

    const columns = [
        {Header: t`Machine Errors`, columns: [
                makeColumn('error', t`Error`, "string"),
                makeColumn('errno', t`Err No.`, "string"),
                {
                    Header: t`Error Type`,
                    id: 'errortype',
                    accessor: 'errortype',
                    filterable: true,
                    Filter: ({ filter, onChange }: { filter: Filter, onChange: ReactTableFunction}) => (
                        <SelectField
                            options={
                                [
                                    {label: "unknown", value: 0},
                                    {label: "info", value: 1},
                                    {label: "alarm", value: 2},
                                    {label: "error", value: 3},
                                    {label: "all", value: ""}
                                ]
                            }
                            onChange={(name, value) => onChange(value)}
                            value={filter ? filter.value : ""}
                        />
                    )
                },
                {
                    Header: t`Status`,
                    id: 'status',
                    accessor: 'status',
                    filterable: true,
                    Filter: ({ filter, onChange }: { filter: Filter, onChange: ReactTableFunction}) => (
                        <SelectField
                            options={
                                [
                                    {label: "raised", value: 0},
                                    {label: "acknowledged", value: 1},
                                    {label: "finished", value: 2},
                                    {label: "all", value: ""}
                                ]
                            }
                            onChange={(name, value) => onChange(value)}
                            value={filter ? filter.value : ""}
                        />
                    )
                },
                {
                    Header: t`Job Number`,
                    id: 'job_number',
                    filterable: true,
                    accessor: (jobState: ErrorModel) => {
                        const job = jobState.job as JobModel;
                        if(!job){
                            return 'none';
                        }

                        return job.number;
                    }
                },
                {
                    Header: t`Date - Time`,
                    id: 'time',
                    accessor: 'time',
                    Cell: (cellInfo: {value: Moment }) => {
                        return getDateTimeString(cellInfo.value);
                    }
                }
            ]
        }];

    const exportCallback = async (startDate: string, endDate: string) => {

        const errors = await machine.exportCSV('errors', startDate, endDate);

        return errors;
    };

    return (
        <MachineDataViewTab
            columns={columns}
            machine={machine}
            enabledKey={'error_enabled'}
            tabName={'Error'}
            store={ErrorStore}
            exportedFileName={'Errors.csv'}
            exportCallback={exportCallback}
        />
    );
});
