import * as React from "react";
import {observer} from "mobx-react-lite";
import {t} from "@lingui/macro";
import {TabPanelProps} from "../TabPanel"
import {MachineModel} from "../../../stores/MachineStore";
import {MachineDataViewTab} from "../MachineDataViewTab";
import {JobStateModel, JobStateStore} from "../../../stores/JobStateStore";
import {JobModel} from "../../../stores/JobStore";
import {Moment} from "moment";
import {getDateTimeString} from "../MachineDataView";
import {ArrowDropDown, ArrowRight} from "@material-ui/icons";
import {ProcessParamsTable} from "../../job_history/ProcessParamsTable";
import {useEffect, useState} from "react";
import {CircularProgress, makeStyles, Typography} from "@material-ui/core";
import {useJobSideColumn} from "../../../hooks/table_columns/useJobSideColumn";

type PanelProps = {
    machine: MachineModel;
} & TabPanelProps

type ProcessParamsData = ReadonlyArray<Readonly<{ name: string | null, value: string | null}>>;

const useStyles = makeStyles((theme) => ({
    processParamTable: {
        "& .rt-tbody": {
            height: 'unset'
        }
    },
    loadingContainer: {
        display: 'flex',
        flexDirection: 'row',
        alignItems: 'center',
        justifyContent: 'center',
        padding: theme.spacing(2)
    },
    loadingText: {
        marginLeft: theme.spacing(1)
    }
}));

export const JobsInfoPanel: React.FunctionComponent<PanelProps> = observer((props) => {
    const {machine} = props;
    const store = JobStateStore;
    const classes = useStyles();

    const columns = [
        {
            Header: t`Jobs`,
            columns: [
                {
                    expander: true,
                    Header: () => null,
                    Expander: ({ isExpanded, ...rest }) => {
                        if(rest.original.state?.toLowerCase() === 'finished') {
                            return(
                                <div>
                                    {isExpanded ? <ArrowDropDown/> : <ArrowRight/>}
                                </div>
                            )
                        } else {
                            return null
                        }
                    },
                },
                {
                    Header: t`Job Number`,
                    id: 'job_number',
                    filterable: true,
                    accessor: (jobState: JobStateModel) => {
                        const job = jobState.job as JobModel;

                        if(!job){
                            return 'n/a';
                        }
                        return job.number;
                    }
                },
                useJobSideColumn(machine, `side`),
                {
                    Header: t`Date - Time`,
                    id: 'time',
                    accessor: 'time',
                    Cell: (cellInfo: {value: Moment }) => {
                        return getDateTimeString(cellInfo.value);
                    }
                }
            ]
        }
    ];

    const processParamsSubComponent = (original: JobStateModel) => {
        if (original.state != 'finished') {
            return null;
        }
        return <ProcessParamsAsync original={original} />;
    }

    const ProcessParamsAsync: React.FunctionComponent<{original: JobStateModel}> = ({original}) => {
        const [processParams, setProcessParams] = useState<ProcessParamsData|null>(null);

        useEffect(() => {
            const fetchData = async () => {
                const unfilteredProcessParams = await original.getProcessParams();
                const filteredProcessParams = unfilteredProcessParams.process_params.map(p => ({
                    name: p.name,
                    value: p.value
                }));
                setProcessParams(filteredProcessParams);
            };
            fetchData();
        }, []);

        if (!processParams) {
            return <div className={classes.loadingContainer}>
                <CircularProgress />
                <Typography className={classes.loadingText} variant="subtitle1">
                    {t`Fetching Data…`}
                </Typography>
            </div>;
        }

        return <div className={classes.processParamTable}>
            <ProcessParamsTable processParams={processParams} />
        </div>;
    };

    const exportCallback = async (startDate: string, endDate: string) => {
        const jobs = await machine.exportCSV('jobs', startDate, endDate);
        return jobs;
    };

    return (
        <MachineDataViewTab
            columns={columns}
            machine={machine}
            enabledKey={'job_state_enabled'}
            tabName={'Job'}
            store={store}
            defaultFiltered={[{id: 'state', value: 5}]}
            subComponent={processParamsSubComponent}
            exportedFileName={'Jobs.csv'}
            exportCallback={exportCallback}
        />
    );
});
