import * as React from "react";
import {observer} from "mobx-react-lite";
import {ReportComponentMetaData, TemplateReportComponent, RangeBarChart} from "@rx-suite/ui";

type MachinePerformanceBatchTimelineReportProps = {
    component: {
        meta: ReportComponentMetaData;
        data: any;
    };
    startTime: string;
    endTime: string;
}
export const MachinePerformanceBatchTimelineReport: React.FunctionComponent<MachinePerformanceBatchTimelineReportProps> = observer(({component, startTime, endTime}) => {
    const data = React.useMemo(() => component.data, [component]);
    const avgDurations = React.useMemo(() => data.series.find(s => s.name === "average_durations"), [data.series]);
    const machineCount = 1;
    const seriesName = (name: string) => {
        if (name === "finished_durations") return "Completed Batches";
        if (name === "aborted_durations") return "Aborted Batches";
        if (name === "ongoing_batch") return "Ongoing Batches";
    }
    const seriesColor = (name: string) => {
        if (name === "finished_durations") return 'rgb(26,93,173)';
        if (name === "aborted_durations") return 'rgb(255,0,0)';
        if (name === "ongoing_batch") return 'rgb(128,128,128)';
    }
    const dataSeries = React.useMemo(() => () => {
        const list: { name: string | undefined; data: { x: string; y: number[]; }[]; color: string | undefined; }[] = []
        // Mapping through aborted and finished data and parse dates as JS Date object
        const filteredSeries = data.series.filter(s => s.name != "average_durations")
        filteredSeries.map((series) => {
            const name = seriesName(series.name);
            const status = name.split(" ")[0]
            const color = seriesColor(series.name);
            const range: { x: string; y: number[]; }[] = [];
            series.data.map((datum: { y: (string | number | Date)[]; x: any; }, index: number) => {
                const y1 = new Date(datum.y[0]).getTime();
                const y2 = new Date(datum.y[1]).getTime();
                const singleMachineLabel = `${status} Batch ${status === "Ongoing" ? "" : (index+1).toString()}`
                const useSingleMachineLabel = machineCount == 1 && Object.keys(avgDurations.data).length > 0
                const xValue = useSingleMachineLabel ? singleMachineLabel : datum.x
                if (useSingleMachineLabel && y1 && y2 || !useSingleMachineLabel) {
                    range.push({x: xValue, y: [y1, y2]})
                }
            })
            list.push({
                name: name,
                data: range,
                color: color
            })
        })
        return list
    }, [data]);
    const avgValueFormatter = React.useMemo(() => (value: string, index: any) => {
        if (dataSeries().some(x => x.data.length > 0)) {
            if (machineCount > 1 && avgDurations.data[value]) {
                return [value, `avg. ${avgDurations.data[value]} mins`]
            } else {
                return value
            }
        } else {
            return ''
        }
    }, [avgDurations.data])

    const titleText = React.useMemo(() => () => {
        if (machineCount > 1 || Object.keys(avgDurations.data).length == 0) {
            return "Batches Duration"
        } else {
            return ["Batches Duration", `avg. ${Object.values(avgDurations.data)[0]} mins`]
        }
    }, [machineCount])

    return(
        <TemplateReportComponent meta={component.meta}>
            <RangeBarChart series={dataSeries()}
                           options={{
                               title: {
                                   text: titleText(),
                                   align: "center"
                               },
                               chart: {
                                   toolbar: {
                                       show: false
                                   },
                                   zoom: {
                                       enabled: false,
                                   }
                               },
                               plotOptions: {
                                   bar: {
                                       barHeight: "98%",
                                       horizontal: true,
                                       dataLabels: {
                                           hideOverflowingLabels: false
                                       }
                                   }
                               },
                               xaxis: {
                                   type: 'datetime',
                                   min: new Date(`${startTime.split('-', 3).join("-")}Z`).getTime(),
                                   max: new Date(`${endTime.split('-', 3).join("-")}Z`).getTime(),
                               },
                               yaxis: {
                                   labels: {
                                       formatter: avgValueFormatter
                                   }
                               },
                               legend: {
                                   position: 'bottom'
                               },
                               fill: {
                                   type: 'solid',
                                   opacity: 0.6
                               },
                               tooltip: {
                                   x: {
                                       format: 'yy-MM-dd HH:mm'
                                   }
                               }
                           }}
                           height={'100%'}
            />
        </TemplateReportComponent>
    )
})