import * as React from 'react';
import {observer} from "mobx-react-lite";
import {LineChart, ReportComponentMetaData, TemplateReportComponent} from "@rx-suite/ui";

type MachinePerformanceReportProps = {
    component: {
        meta: ReportComponentMetaData;
        data: any;
    };
};

export const MachinePerformanceReport: React.FunctionComponent<MachinePerformanceReportProps> = observer(({component}) => {
    const data = component.data;

    const colors = {
        uptime: {
            over: 'rgb(215,228,189)',
            under: 'rgb(228,202,189)'
        },
        utilization: {
            over: 'rgb(146,208,80)',
            under: 'rgb(208,103,80)'
        },
        throughput: 'rgb(26,93,173)'
    }

    const seriesNames = (name: string) => {
        if(name === 'waitingforjobs'){
            return 'uptime';
        }
        if(name === 'processing'){
            return 'utilization';
        }
        return name;
    }
    const UPTIME_SERIES_INDEX = 1;
    const UTILIZATION_SERIES_INDEX = 0;

    const uptimeValueForDataPoint = React.useCallback((dataPointIndex: number) => {
        // sum the uptime and the processing of the series to get the utilization
        return data.series[0]['data'][dataPointIndex].y + data.series[1]['data'][dataPointIndex].y;
    },[data.series])


    const seriesPointValue = (value, {seriesIndex, dataPointIndex}) => {
        if(seriesIndex === UPTIME_SERIES_INDEX){
            value = uptimeValueForDataPoint(dataPointIndex);
        }

        if(seriesIndex == UTILIZATION_SERIES_INDEX || seriesIndex == UPTIME_SERIES_INDEX){
            return `${value.toFixed(2)}%`;
        }

        return `${value} jobs`;
    }
    return (
        <TemplateReportComponent meta={component.meta}>
            <LineChart series={data.series}
                       options={
                           {
                               title: {
                                   text: 'Performance Overview',
                                   align: 'center'
                               },
                               chart: {
                                   stacked: true,
                                   toolbar: {
                                       show: false
                                   },
                                   zoom: {
                                       enabled: false
                                   }
                               },
                               tooltip: {
                                   y: {
                                       formatter: seriesPointValue,
                                       title: {
                                           formatter: seriesNames
                                       },
                                   },
                                   fillSeriesColor: false,
                                   marker: false
                               },
                               legend: {
                                   position: 'bottom',
                                   markers: {
                                       fillColors: [colors.utilization.over, colors.uptime.over, colors.throughput]
                                   },
                                   formatter: seriesNames
                               },
                               xaxis: {
                                   type: 'category'
                               },
                               dataLabels: {
                                   enabled: true,
                                   enabledOnSeries: [2]
                               },
                               stroke: {
                                   width: 2
                               },
                               yaxis: [
                                   {
                                       min: 0,
                                       max: 100,
                                       floating: false,
                                       tickAmount: 5,
                                       forceNiceScale: false,
                                       decimalsInFloat: 0,
                                       seriesName: 'waitingforjobs',
                                       title: {
                                           text: 'Utilization/Uptime (%)'
                                       },
                                   },
                                   {
                                       show: false,
                                       min: 0,
                                       max: 100,
                                       floating: false,
                                       tickAmount: 5,
                                       seriesName: 'processing',
                                       forceNiceScale: false,
                                   },
                                   {
                                       seriesName: 'throughput',
                                       opposite: true,
                                       tickAmount: 5,
                                       decimalsInFloat: 0,
                                       title: {
                                           text: 'Throughput (jobs)'
                                       }
                                   }
                               ],
                               colors: [
                                   function({ value, dataPointIndex}) {
                                       const target = data.options.targets.utilization ? data.options.targets.utilization : 0;

                                       if (value >= target) {
                                           return colors.utilization.over;
                                       } else {
                                           return colors.utilization.under;
                                       }
                                   },
                                   function({ dataPointIndex }) {
                                       const value = uptimeValueForDataPoint(dataPointIndex)
                                       const target = data.options.targets.utilization ? data.options.targets.uptime : 0;

                                       if (value >= target) {
                                           return colors.uptime.over;
                                       } else {
                                           return colors.uptime.under;
                                       }
                                   },
                                   'rgb(26,93,173)'
                               ]
                           }
                       }
                       height={'100%'}
            />
        </TemplateReportComponent>
    )
})