module EerValueCalculator

  def calculate_eer_value(machines, shifts, date)
    y_arr = []
    machines.each do |machine|
      machine_id = machine.id
      time_loss = get_time_loss(machine_id, shifts)
      loading_loss = get_loading_loss(machine_id, shifts)
      quality_loss = get_quality_loss(machine_id, date)
      # Each of the loss values are between [0, 1], normalize the total to also be between [0, 1]
      normalized_total_loss = (time_loss / 3) + (loading_loss / 3) + (quality_loss / 3)
      y_arr.push(1 - normalized_total_loss)
    end
    return if y_arr.length == 0
    # calculate the average for this point of the series (1 or many machines)
    ((y_arr.sum(0.0) / y_arr.length) * 100.0).round(2)
  end

  def get_shifts(start_date, end_date)
    start_of_range = start_date.to_date.beginning_of_day
    end_of_range = end_date.to_date.beginning_of_day + 1.day # using end_of_day results in a missing second with time loss
    all_shifts = RxModule::Communicator::Shifts.active_ranges_between(start_of_range, end_of_range)
    formatted_shifts = {}
    date_range_labels = Tools.create_range_labels(start_date.to_time, end_date.to_time, '1d')
    date_range_labels.each do |date|
      daily_shift = filter_shifts_for_date(all_shifts, date)
      formatted_shifts[date] = daily_shift
    end
    formatted_shifts
  end

  private

  def filter_shifts_for_date(shifts, date)
    start_of_date = date.to_date.beginning_of_day.utc.iso8601
    end_of_date = (date.to_date.beginning_of_day + 1.day).utc.iso8601
    ranges = []
    active_seconds = 0
    shifts[:ranges].each do |shift|
      start_of_shift = shift[:start]
      end_of_shift = shift[:end]
      next unless start_of_shift.between?(start_of_date, end_of_date) || end_of_shift.between?(start_of_date, end_of_date)
      start_of_shift = start_of_shift.to_time < start_of_date.to_time ? start_of_date : start_of_shift
      end_of_shift = end_of_shift.to_time > end_of_date.to_time ? end_of_date : end_of_shift
      ranges.push({
                                      start: start_of_shift,
                                      end: end_of_shift
                                    })
      active_seconds += end_of_shift.to_time.to_i - start_of_shift.to_time.to_i
    end
    {
      ranges: ranges,
      start: start_of_date,
      end: end_of_date,
      active_seconds: active_seconds
    }
  end

  def get_planned_production_time(machine_id, shifts)
    eer_shifts = shifts[:active_seconds]
    eng_lockout = 0
    shifts[:ranges].each do |shift|
      shift_start = shift[:start]
      shift_end = shift[:end]
      engineering_lockout_events = get_engineering_lockout_events(machine_id, shift_start, shift_end)
      engineering_lockout_events.each do |event|
        next if event[:start_time].nil? || event[:resolved_time].nil?
        # edge case: an event started before the shift, but ends during it
        start_time_in_period = [event[:start_time].to_time, shift_start.to_time].max
        # edge case: an event started during the shift, but ends after it
        end_time_in_period = [event[:resolved_time].to_time, shift_end.to_time].min
        eng_lockout += end_time_in_period - start_time_in_period
      end
    end
    eer_shifts - eng_lockout
  end

  def get_engineering_lockout_events(machine_id, start_time, end_time)
    EerEvent.select(:start_time, :resolved_time)
      .where(event_type: 'engineering_lockout', machine_id: machine_id)
      .where(EerEvent.arel_table[:start_time].gteq(start_time)
                    .and(EerEvent.arel_table[:start_time].lt(end_time))
                    .or(
                      # for engineering_lockout, resolved_time == closed time
                      EerEvent.arel_table[:resolved_time].gteq(start_time)
                              .and(EerEvent.arel_table[:resolved_time].lt(end_time))
                    )
    )
  end

  def get_time_loss(machine_id, shifts)
    prod_time = get_planned_production_time(machine_id, shifts)
    return 0 if prod_time <= 0 # if there is no production time, there is no time loss
    total_time_over_max = 0
    eer_and_shift_looping(machine_id, shifts, [:time_over_max, :start_time, :resolved_time, :max_time]) do |eer_events, shift_start, shift_end|
      time_arr = eer_events.each.with_index.map do |eer_event, i|
        time_over_max = eer_event.time_over_max
        next nil if time_over_max.nil? || time_over_max <= 0
        next_eer_event = eer_events[i + 1]
        next nil if next_eer_event.nil? # last event of array is only needed as the end of the range of the previous event
        # find all engineering lockouts that collide with the event
        engineering_lockout_events = get_engineering_lockout_events(machine_id, eer_event[:start_time].to_time, eer_event[:resolved_time].to_time)
        # if the event start_time is during an engineering lockout, return nil, these events fully overlap with the lockout
        next nil if engineering_lockout_events.any? { |ev| eer_event[:start_time].to_time.between?(ev[:start_time].to_time, ev[:resolved_time].to_time) }
        # else, remove the engineering lockout time from the event time_over_max, to a minimum of 0
        event_data = get_time_over_max_period(eer_event, shift_start, shift_end)
        overlapping_seconds = get_overlapping_seconds(event_data[:start_time].to_time, event_data[:end_time].to_time, engineering_lockout_events)
        real_time_over_max = event_data[:time_over_max] - overlapping_seconds
        real_time_over_max <= 0 ? 0 : real_time_over_max
      end
      total_time_over_max += time_arr.compact.sum(0.0)
    end
    time_loss = total_time_over_max.to_f / prod_time.to_f
    time_loss > 1 ? 1.0 : time_loss # in the case of a rounding issue
  end

  def get_time_over_max_period(eer_event, period_start, period_end)
    max_time = eer_event[:max_time]
    max_time = 0 if max_time.nil? || max_time < 0
    over_max_timestamp = eer_event[:start_time].to_time + max_time
    start_time_in_period = [over_max_timestamp, period_start].max
    end_time_in_period = [eer_event[:resolved_time].to_time, period_end].min
    {
      start_time: start_time_in_period,
      end_time: end_time_in_period,
      time_over_max: [eer_event[:time_over_max], end_time_in_period - start_time_in_period].min
    }
  end

  def get_overlapping_seconds(start_time, end_time, overlap_events)
    total_overlap_time = 0
    # NOTE: the overlap events do NOT overlap themselves
    overlap_events.each do |overlap_event|
      next if overlap_event[:start_time].to_time > end_time
      next if overlap_event[:resolved_time].to_time < start_time
      overlap_start = [start_time, overlap_event[:start_time].to_time].max
      overlap_end = [end_time, overlap_event[:resolved_time].to_time].min
      total_overlap_time += overlap_end - overlap_start
    end
    total_overlap_time
  end

  def get_loading_loss(machine_id, shifts)
    total_lens_count = 0
    total_empty_holes = 0
    eer_and_shift_looping(machine_id, shifts, [:lens_number, :empty_holes]) do |eer_events, shift_start, shift_end|
      eer_events.each_with_index do |eer_event, i|
        lens_count = eer_event.lens_number || 0
        empty_holes = eer_event.empty_holes || 0
        next if lens_count == 0 && empty_holes == 0
        next_eer_event = eer_events[i + 1]
        next nil if next_eer_event.nil? # last event of array is only needed as the end of the range of the previous event
        total_lens_count += lens_count
        total_empty_holes += empty_holes
      end
    end
    return 0 if total_lens_count == 0 && total_empty_holes == 0
    total_empty_holes.to_f / (total_lens_count + total_empty_holes).to_f # calculate loss percent
  end

  def eer_and_shift_looping(machine_id, shifts, filter)
    shifts[:ranges].each do |shift|
      shift_start = shift[:start].to_time
      shift_end = shift[:end].to_time
      eer_events = get_door_events_for_machine(machine_id, filter, shift_start, shift_end)
      yield(eer_events, shift_start, shift_end)
    end
  end

  def get_quality_loss(machine_id, date)
    machine_yield = EerYield.where(machine_id: machine_id)
                            .where(EerYield.arel_table[:date].gteq(date.to_date.beginning_of_day)
                                           .and(EerYield.arel_table[:date].lteq(date.to_date.end_of_day))
                            ).first
    return 0 if machine_yield.nil?
    daily_yield = machine_yield.value > 100.0 ? 100.0 : machine_yield.value
    daily_yield.nil? ? 0 : (1 - (daily_yield / 100.0))
  end

  def get_door_events_for_machine(machine_id, filter, shift_start, shift_end)
    eer_events = all_eer_events(machine_id,
                                filter,
                                sub_query:
                                  EerEvent.arel_table[:start_time].gteq(shift_start)
                                          .and(EerEvent.arel_table[:start_time].lt(shift_end))
    ).order(:start_time).to_a # convert to array so other events can be concatenated
    # edge case: an event can start before the shift
    before_eer_event = all_eer_events(machine_id,
                                        filter,
                                        sub_query: EerEvent.arel_table[:start_time].lt(shift_start)
    ).order(start_time: :desc).first
    eer_events = [before_eer_event] + eer_events if before_eer_event
    # edge case: an event can end after the shift
    after_eer_event = all_eer_events(machine_id,
                                       filter,
                                       sub_query: EerEvent.arel_table[:start_time].gteq(shift_end)
    ).order(:start_time).first
    eer_events = eer_events + [after_eer_event] if after_eer_event
    eer_events
  end

  def all_eer_events(machine_id, *filter, sub_query: '')
    EerEvent.select(:start_time, *filter)
            .where(machine_id: machine_id, event_type: %w[door_open door_closed]).where(sub_query)
  end
end
