module KpiTableHandler
  module Base
    class KpiTableHandler
      # KpiTableHandler is meant to easily group kpi table records into one hour buckets.
      # Data can then be easily bucketed into different timeframes like daily, weekly or monthly.
      # IMPORTANT: KpiTableHandler only works with KpiCacheRecords (30 minute buckets)
      class StartTimeAfterEndTime < StandardError; end

      include KpiHelpers
      include KpiCacheHelpers
      include BucketCalculator
      include ArchiveHelpers
      include IntervalBuilder

      attr_accessor :start_time, :end_time, :timeframe, :machine_ids, :options, :where,  :timeframe_select, :timeframe_group, :db_handler

      def base_select
        raise NotImplementedError
      end

      def base_group
        group = []
        group << table.arel_table[:machine_id] unless aggregate_machines?
        group
      end

      def kpi_to_extract
        raise NotImplementedError
      end

      def table
        raise NotImplementedError
      end

      # Method used to extract the final value out of the KPI cache entry
      def record_value(record)
        raise NotImplementedError
      end

      def aggregate_machines?
        !!options[:aggregate_machines]
      end

      def kpi_table_class
        self.table.klass rescue table
      end

      def machine_count
        machine_ids.size
      end

      def add_time_to_extract
        return false if timeframe_all?
        true
      end

      def use_offset?
        start_time - start_time.beginning_of_hour > 0
      end

      def set_time_frame_grouping_select(select)
        self.timeframe_select = select
      end

      def set_time_frame_grouping_group(group)
        self.timeframe_group = group
      end

      def set_utc_offset_to_time
        @set_utc_offset_to_time = true
      end

      def timeframe_select
        []
      end

      def timeframe_group
        []
      end

      def timeframe_all?
        timeframe == Timeframe::ALL
      end

      # Creates a KpiHandler for the currently used database engine.
      #
      # @ PARAMS
      # start_time        =>    Pull records starting from
      # end_time          =>    Pull records ending at
      # machine_ids       =>    Machines to check records for
      # where             =>    array of filter to apply on the query (ex: [KpiMachineStateDuration.arel_table[:machine_id].eq(2)])
      # group             =>    array of group by (ex: [KpiMachineStateDuration.arel_table[:machine_id]])
      # timeframe         =>    time frame to pull the data from (KpiHelpers::Timeframe)
      #
      # @ RETURN
      # Depending on the options passed and timeframe the kpi_handler will return a different format
      # Timeframe::ALL with non aggregated machines
      # {
      #   machine_id: {
      #     kpis
      #   }
      # }
      # Timeframe::HOUR_BY_HOUR || DAY_BY_DAY || WEEK_BY_WEEK non aggregated machines
      # {
      #   machine_id: {
      #     {
      #       bucket_time: {
      #         kpis
      #       }
      #     }
      #   }
      # }
      #
      # Timeframe::ALL with aggregated_machines
      # {
      #   kpis
      # }
      #
      # Timeframe::HOUR_BY_HOUR || DAY_BY_DAY || WEEK_BY_WEEK aggregated machines
      # {
      #   bucket_time: {
      #     kpis
      #   }
      # }
      def initialize(start_time, end_time, timeframe, machine_ids, options: {}, where: [])
        self.start_time = start_time
        self.end_time = end_time
        self.timeframe = timeframe
        self.machine_ids = machine_ids
        self.options = options

        self.where = where

        self.timeframe_select = []
        self.timeframe_group = []

        if DatabaseServer.mssql?
          self.db_handler = MssqlKpiTableDatabaseHandler.new(use_offset?)
        else
          self.db_handler = PgKpiTableDatabaseHandler.new(use_offset?)
        end

        @intervals = build_time_intervals(start_time, end_time)
        @results = {}
      end

      def run
        append_to_results_helper(build_archive_results) unless @intervals[:archive].nil?

        append_to_results_helper(build_now_results) unless @intervals[:now].nil?

        append_to_results_helper(build_future_results) unless @intervals[:future].nil?

        # Sort hash by machine extended_name (ruby >= 1.9.2 has order preserved hashes)
        keys = @results.keys
        # Make sure keys are machine id and not a string
        return @results unless keys.all? {|k| k.is_a? Integer}
        sorted_machines = Machine.where(id: keys).sort_by {|machine| machine.extended_name}
        @results = sorted_machines.map do |machine|
          [machine.id, @results[machine.id]]
        end.to_h
        @results
      end

      def timeframe_boundary
        {
          Timeframe::HOUR_BY_HOUR => {method: :beginning_of_hour, args: nil},
          Timeframe::DAY_BY_DAY => {method: :beginning_of_day, args: nil},
          # We bucket on the day of the week from the first bucket for week by week.
          Timeframe::WEEK_BY_WEEK => {method: :beginning_of_week, args: start_time.strftime('%A').downcase.to_sym}
        }
      end

      def build_query_for_cached_records(interval_start_time, interval_end_time)
        from_sub_query = construct_kpi_table_sub_query(interval_start_time, interval_end_time)

        select_string = add_time_to_extract ? "time, " : ""
        select_string += kpi_to_extract

        query = table.select(select_string).from(from_sub_query)

        # By default, KPI entries are combined to be hourly
        unless timeframe == Timeframe::HOUR_BY_HOUR
          # We are creating a sub table from the query so that we can order by time afterwards instead of ordering by:
          query = db_handler.group_kpi_table_by_time_frame(query, timeframe, timeframe_select, timeframe_group, start_time)
          cols = timeframe_all? ? col_name : [*col_name, :time]
          query = table.select(cols).from(sub_table_to_from_string(query))
        end

        query
      end

      private

      # Returns an sql string of the kpi table grouped by 1 hour intervals
      def construct_kpi_table_sub_query(start_time, end_time)
        # Build the from query specific to the KPI itself
        from_sub_query = db_handler.from_sub_query(table, base_select, base_group)

        # Attach to the from query the where clauses used by all KPI cache entries
        (where + machine_kpi_where(start_time, end_time)).each do |where|
          from_sub_query = from_sub_query.where(where)
        end

        # We need to check for backwards cache building here, since after this point we combine two cache together to build the 1 hour interval.
        # Since we use 2 cache to build the one hour interval, we wont know after this point if one of the cache was missing.
        check_for_backwards_cache(from_sub_query, start_time, end_time)

        sub_table_to_from_string(from_sub_query)
      end

      # Every KPI calculation requires a start time, end time and a set of machine ids
      def machine_kpi_where(start_time, end_time)
        [
          kpi_table_class.arel_table[:time].gteq(start_time),
          kpi_table_class.arel_table[:time].lt(end_time),
          #kpi_table_class.arel_table[:time].lt(end_time+KpiHelpers.cache_entry_size),
          kpi_table_class.arel_table[:machine_id].in(machine_ids)
        ]
      end

      def check_for_backwards_cache(from_sub_query, start_time, end_time)
        # Ensure any cache entry that need to be recalculated (are dirty) are cleared before calculating
        KpiCacheManager.delete_dirty_rows(table, start_time, end_time)

        return if all_cache_entries_exist?(from_sub_query, start_time, end_time)

        find_and_build_missing_cache_entries(from_sub_query, start_time, end_time)
      end

      def interval_start_end_time(target)
        [@intervals[target][:bucket_start_time], @intervals[target][:bucket_end_time]]
      end

      def build_future_results
        interval_start_time, interval_end_time = interval_start_end_time(:future)
        cache = get_outside_cache_buckets(interval_start_time, interval_end_time)

        get_results_from_hash(cache)
      end

      def build_archive_results
        interval_start_time, interval_end_time = interval_start_end_time(:archive)

        cache = {}
        ActiveRecord::Base.connected_to(shard: :archive, role: :reading) do
          cache = get_outside_cache_buckets(interval_start_time, interval_end_time)
        end

        get_results_from_hash(cache)
      end

      def build_now_results
        interval_start_time, interval_end_time = interval_start_end_time(:now)
        query = build_query_for_cached_records(interval_start_time, interval_end_time)

        # Timeframe::ALL only returns one record with no time, so just extract
        if timeframe_all?
          get_results_from_query_all(query)
        else
          # To build the results, the records must be ordered otherwise OFFSET will not work
          query = query.order(:time)
          get_results_from_query(query)
        end
      end

      def get_results_from_hash(cache)
        results = {}
        # Apply the same operation that groups the cache bucket DB side.
        cache.each do |bucket|
          unless aggregate_machines?
            machine_id = bucket[:machine_id]
            results[machine_id] = {} if results[machine_id].nil?
          end

          # Group cache entries by their timeframe (same hour, same day, etc.) and merge their inner values.
          # This is so buckets from different times (future vs now) are added together.
          # ex:
          # 'now' return from 0h00 to 14h30 sept 17 and 'future' return from 14h30 to 23h59 sept 17.
          #  If we are grouping daily, we would have two 'sept 17' buckets (one starting at 0h00 and the other at 14h30) so we merge their value.
          #  To do this, first reduce each bucket to the appropriate 'start time' (i.e start of day) then merge their inner value.
          #
          # Two buckets can have the same key (ex: 14h00 and 14h30) so we have to merge those two buckets to the same timeframe (14h00)

          # Daily with offset is different since we need to backtrack to the previous hour, which could be on a different day/month/year
          time = bucket_time(bucket[:time])

          value = hash_col_select(bucket)

          if aggregate_machines?
            if timeframe_all?
              results.merge!(value) { |k, a, b| a + b}
            else
              results[time] = {} if results[time].nil?
              results[time].merge!(value) { |k, a, b| a + b}
            end
          else
            if timeframe_all?
              results[machine_id].merge!(value) { |k, a, b| a + b}
            else
              results[machine_id][time] = {} if results[machine_id][time].nil?
              results[machine_id][time].merge!(value) { |k, a, b| a + b}
            end
          end
        end

        apply_hash_kpi_operation(results)
      end

      def bucket_time(time)
        return start_time if timeframe_all?

        if timeframe == Timeframe::HOUR_BY_HOUR && use_offset?
          if time.min == 30
            time
          else
            one_hour_less_bucket_time = time - 1.hour
            one_hour_less_bucket_time.change({min: 30})
          end
        else
          boundary = timeframe_boundary[timeframe]
          if boundary[:args].nil?
            time.send(boundary[:method])
          else
            time.send(boundary[:method], boundary[:args])
          end
        end
      end

      # Extract only the columns we want out of the hash (exclude things like machine id, etc.)
      def hash_col_select(bucket)
        bucket.slice(*hash_targets_col)
      end

      def hash_targets_col
        raise NotImplementedError # returns an array of the columns to select out of the hash
      end

      def apply_hash_kpi_operation(hash)
        hash
      end

      def get_results_from_query(query)
        results = {}
        KpiHelpers.get_time_ordered_batches(query) do |batch|
          batch.each do |record|
            extract_result_from_cache_entry(record, results)
          end
        end
        results
      end

      def get_results_from_query_all(query)
        results = {}
        if aggregate_machines?
          # query.take throws error incorrect syntax near offset
          # first needs an order
          # so I think this is the cleanest answer
          results = record_value(query.to_a.first)
        else
          # Cant use batching enumerator methods since theres no ids to order on or a column to order with
          query.each do |record|
            results[record[:machine_id]] = record_value(record)
          end
        end
        results
      end

      def extract_result_from_cache_entry(record, results)
        time = record[:time]
        unless timeframe == Timeframe::HOUR_BY_HOUR
          time = time.advance(seconds: -1 * time.utc_offset) # reverse rails utc offset conversion (Already done db side)
        end
        if aggregate_machines?
          results[time] = record_value(record)
        else
          machine_id = record[:machine_id]
          results[machine_id] = {} if results[machine_id].nil?
          results[machine_id][time] = record_value(record)
        end
      end

      def get_outside_cache_buckets(start_time, end_time)
        cache_entries = []
        get_cache_bucket_times(start_time, end_time).each do |bucket_time|
          machine_ids.each do |machine_id|
            buckets = table.backwards_cache_entry_calculation({time: bucket_time, machine_id: machine_id})
            cache_entries.concat buckets
          end
        end
        cache_entries
      end
    end
  end
end
