module MachineKpiCalculator
  class JobStateDuration < MachineKpiCalculator
    include ArchiveHelpers
    include KpiHelpers
    class << self
      def get_for_period_and_machines(start_time, end_time, timeframe, machine_ids, options: {})
        kpi = JobStateDuration.new(start_time, end_time, timeframe, machine_ids, options)
        kpi.run
      end
    end

    def initialize(start_time, end_time, timeframe, machine_ids, options)
      super(start_time, end_time, timeframe, machine_ids, options)

      @intervals = build_time_intervals(start_time, end_time)
    end

    def build_time_intervals(start_time, end_time)
      {
        archive: archive_times(start_time, end_time),
        now_to_future: now_to_future_times(start_time, end_time)
      }
    end

    def archive_times(start_time, end_time)
      latest_archive_time = data_expired_date + 1.days
      return nil if start_time > latest_archive_time
      {
        start_time: start_time,
        end_time: [end_time, latest_archive_time].min
      }
    end

    def now_to_future_times(start_time, end_time)
      first_now_time = data_expired_date + 1.days
      return nil if end_time < first_now_time
      {
        start_time: [start_time, first_now_time].max,
        end_time: end_time
      }
    end

    def run
      # Get job states from now, future and archive time intervals where applies
      # Only passing the query at this point, so we can load them into batches when building the results
      job_state_queries = []
      job_state_queries << get_archive_job_states unless @intervals[:archive].nil?
      job_state_queries << get_now_to_future_job_states unless @intervals[:now_to_future].nil?

      return build_results(job_state_queries) if aggregate_machines?

      build_results_per_machine(job_state_queries)
    end

    def get_archive_job_states
      interval_start_time, interval_end_time = @intervals[:archive][:start_time], @intervals[:archive][:end_time]
      ActiveRecord::Base.connected_to(shard: :archive, role: :reading) do
        JobState.where(
          time: interval_start_time..interval_end_time,
          machine_id: @machine_ids
        ).order(time: :asc)
      end
    end

    def get_now_to_future_job_states
      interval_start_time, interval_end_time = @intervals[:now_to_future][:start_time], @intervals[:now_to_future][:end_time]
      JobState.where(
        time: interval_start_time..interval_end_time,
        machine_id: @machine_ids
      ).order(time: :asc)
    end

    def build_results(job_state_queries)
      get_state_durations(job_state_queries, false)
    end

    def build_results_per_machine(job_state_queries)
      get_state_durations(job_state_queries, true)
    end

    def get_state_durations(job_state_queries, per_machine = false)
      result = {finished: [], aborted: [], started: [], avg: {}}
      machine_ids.each do |m|
        machine_sn = Machine.find(m).ensured_name
        result[:finished] << {x: machine_sn, y: []}
      end
      durations = []
      last_state_per_machine = {}
      # iterating over the archive query and present query so we can use find_each with default batch size 1000
      job_state_queries.each do |job_state_query|
        KpiHelpers.get_time_ordered_batches(job_state_query) do |batch|
          batch.each_with_index do |job_state, index|
            last_state = last_state_per_machine[job_state.machine_id]
            last_state_as_start = false
            # If started set it as last state for machine and go to next state
            if job_state.state == "started"
              last_state_per_machine[job_state.machine_id] = job_state
              machines_last_job_state = batch.rindex { |x| x.machine_id == job_state.machine_id}
              next if index != machines_last_job_state
              last_state_as_start = true
            end

            if last_state.nil? && !last_state_as_start
              state_end_time = start_time
            elsif last_state_as_start
              state_end_time = end_time
            else
              state_end_time = last_state.time
            end
            duration = if last_state_as_start
                         state_end_time - job_state.time
                       else
                         job_state.time - state_end_time
                       end
            state = job_state.state
            machine_sn = Machine.find(job_state.machine_id).ensured_name
            x_value = per_machine ? machine_sn : (result[state.to_sym].length + 1).to_s
            y_value = if last_state_as_start
                        [job_state.time.strftime("%F %H:%M:%S.%LZ"), state_end_time.strftime("%F %H:%M:%S.%LZ")]
                      else
                        [state_end_time.strftime("%F %H:%M:%S.%LZ"), job_state.time.strftime("%F %H:%M:%S.%LZ")]
                      end
            result[state.to_sym] << {x: x_value, y: y_value}
            # Get average duration per machine or overall duration
            if per_machine && state != "started"
              result[:avg][x_value.to_sym] = [] if !result[:avg][x_value.to_sym]
              result[:avg][x_value.to_sym] << (duration)
            else
              unless state == "started"
                durations << (duration)
              end
            end
            last_state_per_machine[job_state.machine_id] = job_state
          end
        end
      end
      # Get average durations
      if per_machine
        result[:avg].each do |key, value|
          result[:avg][key] = ((value.sum(0.0) / value.size) / 60).round(2) # Avg duration will show in minutes
        end
      else
        result[:avg][:overall] = ((durations.sum(0.0) / value.size) / 60).round(2) # Avg duration will show in minutes
      end

      result
    end
  end
end