# == Schema Information
#
# Table name: block_pickings
#
#  id                   :bigint           not null, primary key
#  block_count          :integer          default(0)
#  delay_on_socket_open :integer          default(0)
#  device_type          :string(4000)
#  hide_block_count     :boolean          default(FALSE)
#  hide_job_number      :boolean          default(FALSE)
#  hide_tray_number     :boolean          default(FALSE)
#  hold_connection      :boolean          default(FALSE)
#  ip_address           :string(4000)
#  name                 :string(4000)
#  port                 :integer
#  scan                 :string(4000)
#  send_job_states      :boolean          default(FALSE)
#  use_scan_count       :boolean          default(FALSE)
#  created_at           :datetime         not null
#  updated_at           :datetime         not null
#  external_dcs_id      :integer
#
class BlockPicking < ApplicationRecord
  include DataHelpers
  include Sideable

  has_many :block_picking_shift_references, :dependent => :destroy
  has_many :shift_references, :through => :block_picking_shift_references
  has_many :block_picking_tags

  class DcsConnectionNotConfigured < StandardError; end

  class << self
    def migrate_legacy_connections
      BlockPicking.where(external_dcs_id: nil).find_each do |conf|
        resp = RxModule::Communicator.call_module_method_as_module("rx_link", "/api/dcs_configs", :post, {
          dcs_config: {
            name: "#{conf.ip_address}:#{conf.port}",
            description:  conf.name,
            ip: conf.ip_address,
            port: conf.port,
            device_type: conf.device_type.nil? ? 'BLK' : conf.device_type,
            retry_on_fail_connection: 3,
            connection_timeout: 5,
            max_open_connections: nil,
            delay_on_socket_open: conf.delay_on_socket_open <= 0 ? nil : conf.delay_on_socket_open,
            hold_connection_open: conf.hold_connection,
            keep_call_trace: false
          }
        })

        puts "Successfully registered DCS #{conf.ip_address}:#{conf.port} with Rx-Link"

        conf.update!(ip_address: nil, port: nil, device_type: nil, delay_on_socket_open: nil, hold_connection: nil, external_dcs_id: resp[:id])
      rescue => err
        puts "Failed to register DCS #{conf.ip_address}:#{conf.port} with Rx-Link: #{err}"
      end
    end
  end

  def get_block_info(tray_number)
    raise DcsConnectionNotConfigured if external_dcs_id.nil?

    resp = BlockPickingRequest.new(external_dcs_id, tray_number, block_picking_tags)

    return resp if resp.error

    increase_block_pick_count(resp.data)
    handle_post_job_states(tray_number, resp.data)

    resp
  end

  def release_count
    self.update(block_count: 0)
  end

  def active_shift?
    ShiftHandler.has_active_shift?(hub_shift_ids)
  end

  def active_shifts
    ShiftHandler.get_active_shifts_for_block_picking(hub_shift_ids)
  end

  def first_active_shift_id
    return nil unless active_shift?
    active_shifts[0][:shift][:id]
  end

  def hub_shift_ids
    self.shift_references.map { |shift| shift.hub_shift_id }
  end

  def handle_post_job_states(tray_number, job_data)
    return unless self.send_job_states
    post_job_state(self.name, tray_number, LensSide::RIGHT, JobStates::FINISHED, process_name) if job_data[:do_right]
    # Cant send them back to back at the same time, this creates a race condition in routing where 2 jobs would be created instead of one.
    post_job_state(self.name, tray_number, LensSide::LEFT, JobStates::FINISHED, process_name, wait: 15.seconds) if job_data[:do_left]
  end

  def increase_block_pick_count(job_data)
    return unless active_shift?

    count = if use_scan_count
              # When use scan count is enabled, we don't care how many lens have been blocked.
              1
            else
              (job_data[:do_left] ? 1 : 0) + (job_data[:do_right] ? 1 : 0)
            end

    self.update(block_count: self.block_count+count)
  end

  def should_broadcast?
    true
  end

  private

  def process_name
    "BLOCK_PICKING"
  end
end
