# == Schema Information
#
# Table name: calibrations
#
#  id               :bigint           not null, primary key
#  calibration_type :string(4000)
#  machine_sn       :string(4000)
#  mode             :string(4000)
#  sub_type         :string(4000)
#  target           :string(4000)
#  target_name      :string(4000)
#  time             :datetime
#  user             :string(4000)
#  created_at       :datetime         not null
#  updated_at       :datetime         not null
#
# Indexes
#
#  index_calibrations_on_machine_sn  (machine_sn)
#
# Foreign Keys
#
#  fk_rails_...  (machine_sn => machines.sn) ON DELETE => cascade
#
class Calibration < ApplicationRecord
  include Pushable
  include ByMachineAndPeriod
  include ForwardMachineData

  belongs_to :machine, foreign_key: :machine_sn, primary_key: :sn, touch: true
  has_many :calibration_corrections
  has_many :calibration_measurements

  def forwarding_data
    {
      target: target,
      target_name: target_name,
      type: calibration_type,
      subtype: sub_type,
      mode: mode,
      user: user,
      measurements: {
        measurement: calibration_measurements.map {|measurement|
          {
            name: measurement.name,
            value: measurement.value
          }
        }
      },
      corrections: {
        correction: calibration_corrections.map {|correction|
          {
            name: correction.name,
            value: correction.value
          }
        }
      }
    }
  end

  def forwarding_endpoint
    {
      request_type: :put,
      url: "calibration/1.0/add"
    }
  end

  def forwarding_params
    {
      sn: machine_sn,
      date: time.strftime("%Y-%m-%d"),
      time: time.strftime("%H:%M:%S,%L")
    }
  end

  def self.delete_before(date)
    data_archive_settings = DataArchiveSetting.first
    max_delete_queue = data_archive_settings.delete_count_calibration
    return 0 if max_delete_queue.nil? || max_delete_queue <= 0
    deleted_objects = 0
    # Delete the calibrations that have existed before the given date
    # Only delete a maximum of `max_delete_queue` calibrations
    ActiveRecord::Base.connected_to(shard: :primary, role: :writing) do
      calibrations = Calibration.where(Calibration.arel_table[:time].lt(date)).limit(max_delete_queue)
      deleted_objects += calibrations.delete_all
    end
    deleted_objects
  end
end
