module KpiHelpers
  extend ActiveSupport::Concern

  class ParamIsNotActiveSupportTime < StandardError; end

  module Timeframe
    ALL = 0
    HOUR_BY_HOUR = 1
    DAY_BY_DAY = 2
    WEEK_BY_WEEK = 3

    TIMEFRAME_TO_ENUM = {all: 0, hour_by_hour: 1, day_by_day: 2, week_by_week: 3}
  end

  class NoKpiTableBlockGiven < StandardError; end
  class BackwardsCacheEntryNotDefined < StandardError; end

  def self.cache_entry_size
    30.minutes
  end

  def self.cache_entry_size_seconds
    1800.0 # 30 minutes
  end

  # Returns the start_time of the bucket that the specified time belongs
  # ex: time is 10h07 => bucket start is 10h00.
  # ex: time is 11h31 => bucket start is 11h30
  def self.kpi_bucket_time(time)
    raise ParamIsNotActiveSupportTime unless time.is_a? ActiveSupport::TimeWithZone
    diff = time - time.beginning_of_hour
    return time.beginning_of_hour + cache_entry_size if diff >= cache_entry_size
    time.beginning_of_hour
  end

  # Returns if the specified time belongs to the bucket that is now
  def self.time_is_mid_hour(time)
    return (time.beginning_of_hour + 30.minutes) == time
  end

  # Returns if the specified time is the hour midpoint
  def self.bucket_is_current_time?(time)
    now_bucket = kpi_bucket_time(Time.zone.now)
    other = kpi_bucket_time(time)
    now_bucket == other
  end

  # Returns the time elapsed in the bucket that the specified time belongs to
  # ex: time is 10h07 => time elapsed is 7 minutes
  # ex: time is 11h31 => time elapsed is 1 minute
  def self.duration_in_bucket(time)
    start_time = kpi_bucket_time(time)
    end_time = start_time + cache_entry_size
    cache_entry_size - (end_time - time)
  end

  # Returns if the specified time belongs to second half hour of the hour
  # ex: time is 10h07 => false
  # ex: time is 11h31 => true
  def self.time_is_in_second_half_hour?(time)
    bucket = kpi_bucket_time(time)
    start_of_hour = bucket.beginning_of_hour

    bucket - start_of_hour > 0
  end

  # Since the 'now' of a machine depends when the cache was last updated, this helper returns the current bucket time
  # For example, if it is now 10h07 but the last machine row update was at 10h05, the bucket size will be 5 minutes.
  def self.machine_state_duration_current_bucket_size
    Time.zone.now - kpi_bucket_time(Time.zone.now)
  end

  def self.get_kpi_cache_entry(model, time, opts={})
    bucket_time = kpi_bucket_time(time)
    model.where(opts).where(model.arel_table[:time].eq(bucket_time)).order("time DESC").first
  end

  def self.time_buckets_between(start_time, end_time, by_hour=false)
    if (end_time - start_time) < 30.minutes
      return [KpiHelpers.kpi_bucket_time(start_time)]
    end
    start_bucket = KpiHelpers.kpi_bucket_time(start_time)
    end_bucket = KpiHelpers.kpi_bucket_time(end_time-30.minutes) # the last time frame is included in the n-1 bucket
    increments = by_hour ? 1.to_f/24 : 1.to_f/48
    start_bucket.to_datetime.step(end_bucket.to_datetime, increments).to_a.map {|b| Time.zone.parse(b.to_s)}
  end

  # find_each (and other rails batches methods) only order by primary key (typically id).
  # In our case we mostly want them ordered by time, so made our own method
  def self.get_time_ordered_batches(query, offset: 0, limit: 1000, &block)
    while offset < query.count(:time)
      res = query.limit(limit).offset(offset)
      block.call(res)
      offset += limit
    end
  end

  def sub_table_to_from_string(sub_table, as=sub_table.table.name)
    "(#{sub_table.to_sql}) as #{as}"
  end

  def aggregate_to_one_value(data)
    return 0 if data.values.empty?
    data.values.sum / data.values.size
  end

  def timeframe_start(time)
    if timeframe == Timeframe::HOUR_BY_HOUR
      time.beginning_of_hour
    else
      time.beginning_of_day
    end

  end

  # Needed for calculating which time bucket a time goes in, like a shift_start time.
  # ex: start_time = 11/11/11 11:00:00, time = 11/11/11 15:35:00, timeframe_length = 1.hour
  # bucket_index = ((15:35:00 - 11:00:00) / 1.hour).floor = 4
  # 11:00:00 + (4 * 1.hour)
  # So time_bucket_index = 15:00:00
  def time_bucket_index(time)
    return timeframe_start(time) if timeframe == Timeframe::ALL

    bucket_index = ((time - start_time) / timeframe_length).floor
    start_time + (bucket_index * timeframe_length)
  end

  def timeframe_length
    case timeframe
    when Timeframe::HOUR_BY_HOUR
      1.hour
    when Timeframe::DAY_BY_DAY
      1.day
    when Timeframe::WEEK_BY_WEEK
      1.week
    when Timeframe::ALL
      end_time - start_time
    end
  end

  def timeframe_all?
    timeframe == Timeframe::ALL
  end
end
