# == Schema Information
#
# Table name: machines
#
#  id                  :bigint           not null, primary key
#  archived            :boolean          default(FALSE)
#  description         :string(4000)
#  is_conveyor_machine :boolean          default(FALSE)
#  name                :string(4000)
#  push_target_enabled :boolean          default(FALSE)
#  sn                  :string(4000)     not null
#  track_eer_metrics   :boolean          default(FALSE), not null
#  created_at          :datetime         not null
#  updated_at          :datetime         not null
#  machine_type_id     :bigint
#
# Indexes
#
#  index_machines_on_machine_type_id  (machine_type_id)
#  index_machines_on_sn               (sn) UNIQUE
#
# Foreign Keys
#
#  fk_rails_...  (machine_type_id => machine_types.id)
#
class Machine < ApplicationRecord
  scope :not_archived, -> { where(archived: false) }
  include Pushable
  include MachineKpi
  include MachineStateDurationHelpers
  include MachineErrorHelpers
  include BoxCoater::MachineHelper
  include ProductionView

  has_many :machine_states, -> { order(:time) }, foreign_key: "machine_sn", primary_key: :sn
  has_many :problems, foreign_key: "machine_sn", primary_key: :sn, class_name: "Error"
  has_many :events, foreign_key: "machine_sn", primary_key: :sn, class_name: "Event"
  has_many :job_processes
  has_many :jobs, through: :job_processes

  has_many :job_states
  has_many :eer_yields
  has_many :kpi_job_states
  has_many :kpi_routing_kickouts

  has_and_belongs_to_many :machine_groups
  has_and_belongs_to_many :process_booking_workstations

  has_many :machine_machine_rejects
  has_many :machine_rejects, class_name: "Machine", through: :machine_machine_rejects

  has_many :machine_reject_machines, class_name: "MachineMachineReject", foreign_key: :machine_reject_id
  has_many :machines, through: :machine_reject_machines

  has_many :box_coater_events
  has_many :box_coater_batches

  belongs_to :machine_type, required: false

  def extended_name
    return sn if name.nil?
    return "#{name} (#{sn})"
  end

  def ensured_name
    name || sn || id
  end

  def current_state
    KpiCurrentMachineState.find_or_create_by(machine_sn: sn)
  end

  def has_machine_type?
    !machine_type_id.nil?
  end

  def belongs_to_group?
    machine_group_ids.size > 0
  end

  #todo: Move to machine state duration KPI helper?
  def set_current_state(machine_state)
    cur_state = current_state
    # If the current machine state is in the future, the machine time sync was off, so replace with newer entry
    if (machine_state.time >= cur_state.time) || (current_state.time >= (Time.zone.now + 15.minutes))
      cur_state.update!(state: machine_state.state, time: machine_state.time, comment: machine_state.comment)
    end
  end

  def self.status_all
    status = {}
    Machine.not_archived.find_each do |m|
      data = m.current_state
      status[m.sn] = {
        machine_id: m.id,
        name: m.name,
        status: data.state,
        comment: data.comment,
        time: data.time,
        duration: Time.now - data.time,
        groups: m.machine_group_ids,
        is_conveyor_machine: m.is_conveyor_machine
      }
    end
    status
  end

  def self.machine_list
    options = []
    other = {label: 'other', options: []}
    Machine.find_each do |machine|
      if machine.has_machine_type?
        type = options.find {|c| c[:label] == machine.machine_type.category}
        if type.nil?
          type = {label: machine.machine_type.category, options: []}
          options << type
        end
        type[:options] << { label: machine.extended_name, value: machine.id }
      else
        other[:options] << { label: machine.extended_name, value: machine.id }
      end
    end
    options << other
    options
  end

  # TODO: this should be retired once 'current_state' is fully implemented
  def last_state
    @last_state ||= machine_states.last
  end

  #TODO: delete?
  reportable("machine", "Machine", "Report on machines", {
    name: {
      name: "Name",
      desc: "The name of the machine"
    },
    sn: {
      name: "Serial Number",
      shortname: "SN",
      desc: "The serial number of the machine"
    },
    description: {
      name: "Description",
      desc: "The description of the machine"
    },
    is_conveyor_machine: {
      name: "Conveyor?",
      desc: "Whether or not the machine is a conveyor"
    },
  })
end
