module ReportComponents
  class MachinePerformanceBatch < ReportComponent
    include DataHelpers
    include ReportComponentHelpers
    include KpiHelpers

    class InvalidTimeFrame < StandardError; end

    def component_id
      'machine_performance_batch'
    end

    def component_type
      timeframe < Timeframe::DAY_BY_DAY  ? 'range-bar' : 'bar'
    end

    def generate(args)
      # Get job state durations within timeframe
      job_state_durations = MachineKpi.job_state_duration_between(start_time, end_time, timeframe, machine_ids, options: {})
      if timeframe > Timeframe::HOUR_BY_HOUR
        average_durations = job_state_durations[:avg].length == 0 ? 0 : (job_state_durations[:avg].values.sum / job_state_durations[:avg].length).round(2)
      else
        average_durations = job_state_durations[:avg]
      end

      series = [
        {
          name: "finished_durations",
          data: job_state_durations[:finished],
          type: 'rangeBar'
        },
        {
          name: "aborted_durations",
          data: job_state_durations[:aborted],
          type: 'rangeBar'
        },
        {
          name: "ongoing_batch",
          data: job_state_durations[:started],
          type: 'rangeBar'
        },
        {
          name: 'average_durations',
          data: average_durations
        }
      ]

      # Job state breakdown is only needed for higher than hour_by_hour
      if timeframe > Timeframe::HOUR_BY_HOUR
        job_state_breakdown = MachineKpi.job_state_breakdown_between(start_time, end_time, timeframe, machine_ids, options: args)
        breakdown_series = []
        job_state_breakdown.map do |machine_id, buckets|
          end_states = %w[finished aborted]
          end_states.map do |state|
            state_in_breakdown = breakdown_series.find { |x| x[:name] == state}
            if state_in_breakdown
              machine_bucket_sum = [state_in_breakdown[:data], buckets.values.map { |b| b[state.to_sym] }].transpose.map {|x| x.reduce(:+)}
              breakdown_series.find { |x| x[:name] == state}[:data] = machine_bucket_sum
            else
              breakdown_series << {
                name: state,
                data: buckets.values.map { |b| b[state.to_sym] },
                color: state == "finished" ? 'rgb(26,93,173)' : 'rgb(255,0,0)',
                type: 'bar'
              }
            end
          end
        end
        # All timestamps are the same, so get one of machine_ids, then get the times from that bucket
        breakdown_labels = job_state_breakdown.values[0].keys.map {|k| bucket_time_to_timeframe_label(k)}
        counts_colors = args[:colorize_series] ? color_for_series(breakdown_series.size - 1) : nil

        # Rearrange values in series to match colors
        even_breakdown = breakdown_series.flatten().select.each_with_index {|_, i| i.even?}
        odd_breakdown =  breakdown_series.flatten().select.each_with_index {|_, i| i.odd?}
        final_breakdown = even_breakdown + odd_breakdown

        # Add job state breakdown to series as "counts"
        series << {
          name: "counts",
          value: final_breakdown,
          labels: breakdown_labels,
          colors: counts_colors
        }
      end

      {
        options: {},
        series: series
      }
    end

    def format_finished(bucket_value)
      bucket_value[:finished]
    end

    def format_aborted(bucket_value)
      bucket_value[:aborted]
    end

    private

    def extract_target_job_state(data, target_state)
      new_data = {}
      data.each do |time_bucket, data|
        new_data[time_bucket] = data.slice(target_state)
      end
      new_data
    end
  end
end

