module ReportComponents
    class MachineStateBreakdownPie < ReportComponent
      include DataHelpers
      include KpiHelpers

      attr_accessor :states, :colors

      def component_id
        'machine_state_breakdown'
      end

      def component_type
        'pie'
      end

      def generate(args)
        @states = MachineKpi.state_breakdown_between(start_time, end_time, Timeframe::ALL, machine_ids, options: {:as_percent => true})

        # Preload all the colors as we'll need them all regardless
        @colors = Color.all.pluck(:status_name, :color)

        series, labels, colors = if ProductionSetting.use_simple_machine_state_colors?
                                   build_simplified_graph_data
                                 else
                                   build_graph_data
                                 end

        series, labels, colors = ensure_data(series, labels, colors)

        return set_no_data('No machine states during reporting period') if series.blank?

        {
          series: series,
          labels: labels,
          colors: colors
        }
      end

      def ensure_data(series, labels, colors)
        return [[], [], []] if series.blank? && labels.blank? && colors.blank?
        series, labels, colors = series.zip(labels, colors).sort.reverse!.transpose
        data_sum = series.sum

        # avoid division by zero
        # don't pad a margin of error
        if data_sum > 0 && (100 - data_sum > 1)
          series << (100 - data_sum)
          labels << 'notconnected'
          colors << '#ABB2B9'
        end
        ## NOTE: At the moment, if there is unresolved data, it is not displayed, instead, the existing data is normalized.
        ## series = series.map { |val| ((val * 100) / data_sum).round(2)} if (data_sum > 0 && (100 - data_sum > 1))
        [series, labels, colors]
      end

      def build_graph_data
        series = []
        labels = []
        colors = []

        @states.each do |state, value|
          value = value.to_f
          next unless value > 0

          series << value
          labels << state
          colors << color_for_state(state)
        end

        [series, labels, colors]
      end

      def build_simplified_graph_data
        series = []
        labels = []
        colors = []

        groups = [
          [:processing],
          [:waitingforjobs],
          [:estop, :halt, :notready, :error],
          [:calibration, :maintenance, :on, :setup, :ready],
          [:off]
        ]

        groups.each do |group|
          value = value_for_machine_state_group(group)
          next unless value > 0

          series << value
          labels << group.join(" - ")
          colors << color_for_state(group[0])
        end

        [series, labels, colors]
      end


      # Only include label if at least one machine state is present and it's value is greater than 0
      def value_for_machine_state_group(states)
        acc = 0
        states.each do |state|
          acc += @states[state] unless @states[state].nil?
        end
        acc.to_f
      end

      def color_for_state(state)
        color = @colors.find{|c| c[0].downcase == state.to_s.downcase }
        if color.nil?
          '#ABB2B9'
        else
          color[1]
        end
      end
    end
end

